/*******************************************************************************
 * Copyright (c) 2006-2007 IONA Technologies.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IONA Technologies - initial API and implementation
 *******************************************************************************/
package org.eclipse.stp.ui.xef.editor;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.IContentProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.Window;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.stp.policy.common.editors.IPolicyDetailCallback;
import org.eclipse.stp.policy.common.editors.IPolicyDetailEditorInput;
import org.eclipse.stp.ui.xef.XefPlugin;
import org.eclipse.stp.ui.xef.schema.IContextProvider;
import org.eclipse.stp.ui.xef.schema.SchemaRegistry;
import org.eclipse.stp.xef.DocumentNamespaceContext;
import org.eclipse.stp.xef.ISchemaProvider;
import org.eclipse.stp.xef.URLSchemaProvider;
import org.eclipse.stp.xef.XMLSnippet;
import org.eclipse.stp.xef.XPathXMLProvider;
import org.eclipse.stp.xef.XefConstants;
import org.eclipse.stp.xef.util.InputStreamHelper;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWizard;
import org.eclipse.ui.dialogs.WizardNewFileCreationPage;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.ui.wizards.newresource.BasicNewFileResourceWizard;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.bootstrap.DOMImplementationRegistry;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSParser;
import org.w3c.dom.ls.LSSerializer;

public class XMLXPathView extends ViewPart implements ISelectionListener {
    private Label fileLabel;
    private Label schemasLabel;
    private Combo xpath;
    private IFile selectedFile;
    private IStructuredSelection selection;
    private TableViewer tableViewer;
    
    static {
        System.setProperty(DOMImplementationRegistry.PROPERTY, "org.apache.xerces.dom.DOMXSImplementationSourceImpl");
    }
    
	public void createPartControl(Composite parent) {
        FormToolkit toolkit = new FormToolkit(parent.getDisplay());
        ScrolledForm form = toolkit.createScrolledForm(parent);
        form.setText("Select portion of file to edit");
        GridLayout layout = new GridLayout(3, false);
        form.getBody().setLayout(layout);
        toolkit.createLabel(form.getBody(), "File:");
        fileLabel = toolkit.createLabel(form.getBody(), "");
        fileLabel.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        toolkit.createLabel(form.getBody(), "");
        toolkit.createLabel(form.getBody(), "Schema Location:");
        schemasLabel = toolkit.createLabel(form.getBody(), "");
        schemasLabel.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        toolkit.createLabel(form.getBody(), "");
        toolkit.createLabel(form.getBody(), "XPath:");
        xpath = new Combo(form.getBody(), SWT.DROP_DOWN);
        toolkit.adapt(xpath);
        xpath.setItems(new String [] {
            "//wsp:All",
            "//*[local-name()='Policy' and namespace-uri()='http://www.w3.org/ns/ws-policy']"});
        xpath.select(0);
        xpath.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));

        Button edit = toolkit.createButton(form.getBody(), "Edit", SWT.PUSH);
        edit.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                try {
                    openEditor(xpath.getText());
                } catch (Exception e1) {
                    e1.printStackTrace();
                }
            }
        });
        
        toolkit.createLabel(form.getBody(), "");
        
        Composite buttons = toolkit.createComposite(form.getBody()); 
        buttons.setLayout(new RowLayout());
        GridData gd = new GridData();
        gd.horizontalSpan = 2;
        buttons.setLayoutData(gd);

        Button test = toolkit.createButton(buttons, "Test XPath", SWT.PUSH);
        test.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                testXPath(true);
            }            
        });
        
        Button create = toolkit.createButton(buttons, "Create empty WS-Policy File", SWT.PUSH);
        create.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent se) {
                IWorkbenchWizard wizard = new BasicNewFileResourceWizard() {

                    @Override
                    public boolean performFinish() {
                        IFile file = ((WizardNewFileCreationPage) getPages()[0]).createNewFile();                        
                        if (file == null) {
                            return false;
                        }

                        String contents = 
                            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + 
                            System.getProperty("line.separator") + 
                            "<wsp:Policy xmlns:wsp=\"http://www.w3.org/ns/ws-policy\" >" +
                            System.getProperty("line.separator") + 
                                "<wsp:All />" + 
                                System.getProperty("line.separator") + 
                            "</wsp:Policy>";
                        try {
                            file.setContents(new ByteArrayInputStream(contents.getBytes()), false, false, new NullProgressMonitor());
                            file.getProject().refreshLocal(IProject.DEPTH_INFINITE, new NullProgressMonitor());
                        } catch (CoreException e) {
                            e.printStackTrace();
                            return false;
                        }
                        selectAndReveal(file);
                        selectionChanged(null, new StructuredSelection(file));

                        // Open editor on new file.
                        IWorkbenchWindow dw = getWorkbench().getActiveWorkbenchWindow();
                        try {
                            if (dw != null) {
                                IWorkbenchPage page = dw.getActivePage();
                                if (page != null) {
                                    xpath.setText("//wsp:All");
                                    openEditor("//wsp:All");
                                }
                            }
                        } catch (Exception e) {
                            IStatus st = new Status(IStatus.ERROR, XefPlugin.ID, 0, "Problems Opening Editor", e);
                            XefPlugin.getDefault().getLog().log(st);
                        }
                        return true;
                    }
                    
                };
                                
                IStructuredSelection sel = selection;
                if (sel == null) {
                    sel = new StructuredSelection(ResourcesPlugin.getWorkspace().getRoot().getProjects());
                }
                
                wizard.init(getSite().getWorkbenchWindow().getWorkbench(), sel);
                WizardDialog dialog = new WizardDialog(getSite().getShell(), wizard);
                wizard.setContainer(dialog);
                dialog.open();
            }            
        });
        
        tableViewer = new TableViewer(form.getBody());
        tableViewer.setContentProvider(getPropertiesContentProvider());
        tableViewer.getTable().setHeaderVisible(true);
        tableViewer.getTable().setLinesVisible(true);
        gd = new GridData(GridData.FILL_BOTH);
        gd.horizontalSpan = 2;
        tableViewer.getControl().setLayoutData(gd);
        
        TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
        column.getColumn().setWidth(250);
        column.getColumn().setText("Setting");
        column.getColumn().setMoveable(true);
        column.getColumn().setResizable(true);
        column.setLabelProvider(new ColumnLabelProvider() {
            @Override
            @SuppressWarnings("unchecked")
            public String getText(Object element) {
                if (element instanceof Map.Entry) {
                    Map.Entry<String, String> entry = (Map.Entry<String, String>) element;
                    return entry.getKey();
                } else {
                    return element.toString(); 
                }
             }
        });

        column = new TableViewerColumn(tableViewer, SWT.NONE);
        column.getColumn().setWidth(250);
        column.getColumn().setText("Value");
        column.getColumn().setMoveable(true);
        column.getColumn().setResizable(true);
        column.setLabelProvider(new ColumnLabelProvider() {
            @Override
            @SuppressWarnings("unchecked")
            public String getText(Object element) {
                if (element instanceof Map.Entry) {
                    Map.Entry<String, String> entry = (Map.Entry<String, String>) element;
                    return entry.getValue();
                } else {
                    return element.toString(); 
                }
             }
        });
        column.setEditingSupport(new AbstractEditingSupport(tableViewer) {
            @Override
            @SuppressWarnings("unchecked")
            protected void doSetValue(Object element, Object value) {
                Map.Entry<String, String> entry = (Map.Entry<String, String>) element;
                entry.setValue(value.toString());
            }

            @Override
            @SuppressWarnings("unchecked")
            protected Object getValue(Object element) {
                Map.Entry<String, String> entry = (Map.Entry<String, String>) element;
                return entry.getValue();                
            }            
        });
                
        final Map<String, String> m = new HashMap<String, String>();
        m.put("TITLE", "Testing Policies");
        tableViewer.setInput(m);
        
        Composite propButtons = toolkit.createComposite(form.getBody()); 
        propButtons.setLayout(new GridLayout(1, true));
        gd = new GridData(GridData.VERTICAL_ALIGN_BEGINNING);
        propButtons.setLayoutData(gd);
        
        Button addBtn = toolkit.createButton(propButtons, "+", SWT.PUSH);
        addBtn.setText("+");
        addBtn.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                SettingsDialog sd = new SettingsDialog(getViewSite().getShell());
                if (sd.open() == Window.OK) {
                    Setting s = sd.getSetting();
                    if (s != null) {
                        m.put(s.toString(), s.getDefault());
                    }
                }
                
                tableViewer.refresh();                
            }            
        });
        gd = new GridData(GridData.VERTICAL_ALIGN_BEGINNING | GridData.FILL_HORIZONTAL);
        addBtn.setLayoutData(gd);
        
        Button removeBtn = toolkit.createButton(propButtons, "-", SWT.PUSH);
        removeBtn.addSelectionListener(new SelectionAdapter() {
            @SuppressWarnings("unchecked")
            @Override
            public void widgetSelected(SelectionEvent e) {
                if (tableViewer.getSelection() != null && !tableViewer.getSelection().isEmpty()) {
                    ISelection s = tableViewer.getSelection();
                    if (s instanceof IStructuredSelection) {
                        Object o = ((IStructuredSelection) s).getFirstElement();
                        if (o instanceof Map.Entry) {
                            m.remove(((Map.Entry) o).getKey());
                            tableViewer.refresh();
                        }
                    }                    
                } else {                    
                    MessageDialog.openInformation(getViewSite().getShell(), "Remove property", "Please select a property in the table to remove"); 
                }
            }            
        });
        gd = new GridData(GridData.VERTICAL_ALIGN_BEGINNING | GridData.FILL_HORIZONTAL);
        removeBtn.setLayoutData(gd);
        
        
        toolkit.paintBordersFor(form.getBody());
        // add myself as a global selection listener
        getSite().getPage().addSelectionListener(this);

        // prime the selection if there is one
        ISelection selection = getSite().getPage().getSelection();
        if (selection != null) {
            selectionChanged(null, selection);
        }        
	}

    @SuppressWarnings("unchecked")
    private void openEditor(String xpath) throws Exception {
        ISchemaProvider sp = getSchemaProvider();

        Map<Setting, String> settings = new HashMap<Setting, String>();
        for (Map.Entry entry : (Set<Map.Entry>) ((Map) tableViewer.getInput()).entrySet()) {
            settings.put(Setting.valueOf(entry.getKey().toString()), entry.getValue().toString());
        }
        
        IContextProvider ctxProvider = new IContextProvider() {
            public Object getData(String ctxId) {
                return null;
            }

            public String[] getValues(String ctxId, String ctxFilter) {
                if ("loglevels".equals(ctxId)) {
                    return new String [] {"Boring", "Interesting"};
                }
                return null;
            }            
        };
        
        XPathXMLProvider xpxp = new XPathXMLProvider(selectedFile, xpath, testXPath(false));
        ByteArrayInputStream bais = new ByteArrayInputStream(xpxp.getXML().getBytes()); 
        IPolicyDetailEditorInput ei = new MyPolicyDetailEditorInput(selectedFile, sp, bais);
        ei.getProperties().put("TITLE", "My Title");
        ei.getProperties().put("HEADING", "My Heading");
        IDE.openEditor(getSite().getWorkbenchWindow().getActivePage(), ei, XefConstants.XEF_EDITOR);
        
/*         IDE.openEditor(getSite().getWorkbenchWindow().getActivePage(),
                new XMLProviderEditorInput(settings, selectedFile.getProject(),
                new XPathXMLProvider(selectedFile, xpath, testXPath(false)), sp, ctxProvider),
                XefConstants.XEF_EDITOR);
                */
    }            

    private IContentProvider getPropertiesContentProvider() {
	    return new IStructuredContentProvider() {

            public Object[] getElements(Object inputElement) {
                if (inputElement instanceof Map) {
                    return ((Map) inputElement).entrySet().toArray();
                }
                return null;
            }

            public void dispose() {}
            public void inputChanged(Viewer viewer, Object oldInput,
                    Object newInput) {}
	        
	    };
    }

	protected abstract class AbstractEditingSupport extends EditingSupport {
        private TextCellEditor editor;

        public AbstractEditingSupport(TableViewer viewer) {
            super(viewer);
            this.editor = new TextCellEditor(viewer.getTable());
        }

        protected boolean canEdit(Object element) {
            return true;
        }

        protected CellEditor getCellEditor(Object element) {
            return editor;
        }

        protected void setValue(Object element, Object value) {
            doSetValue(element, value);
            getViewer().update(element, null);
        }

        protected abstract void doSetValue(Object element, Object value);
    }
	
    @Override
    public void dispose() {
        super.dispose();
        getSite().getPage().addSelectionListener(this);
    }

    protected ISchemaProvider getSchemaProvider() throws Exception {
        File [] files = selectedFile.getRawLocation().toFile().getParentFile().listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return !name.startsWith(".") && name.endsWith(".xsd");
            }            
        });
        
        URL [] urls = new URL[files.length];
        for (int i=0; i < urls.length; i++) {
            urls[i] = files[i].toURI().toURL();
        }
        
        File [] snippetFiles = selectedFile.getRawLocation().toFile().getParentFile().listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.endsWith(".sxml");
            }            
        });
        XMLSnippet [] snippets = new XMLSnippet[snippetFiles.length];
        for (int i=0; i < snippets.length; i++) {
            InputStream fs = new FileInputStream(snippetFiles[i]);
            try {
                snippets[i] = new XMLSnippet(snippetFiles[i].getName(), 
                    new String(InputStreamHelper.drain(fs)));
            } finally {
                InputStreamHelper.close(fs);
            }
        }
        
        // Flush the schema registry every time. Not advisable in production
        // scenarios, as reloading might be expensive. Useful however in testing 
        // scenarios.
        SchemaRegistry.getDefault().clear();
        
        return new URLSchemaProvider(urls, snippets);
    }

    /**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		xpath.setFocus();
	}

    public void selectionChanged(IWorkbenchPart part, ISelection sel) {        
        if (!sel.isEmpty() && sel instanceof IStructuredSelection) {
            selection = (IStructuredSelection) sel;
            Object o = selection.getFirstElement();
            
            if (o instanceof IFile) {
                selectedFile = (IFile) o;
                fileLabel.setText(selectedFile.getRawLocation().toFile().toString());
                schemasLabel.setText(selectedFile.getRawLocation().toFile().getParent() + File.separator + "*.xsd");
            }
        }
    }
    
    private String testXPath(boolean show) {
        String qname = null;
        try {
            DOMImplementationRegistry registry = DOMImplementationRegistry.newInstance();
            DOMImplementationLS domLS = (DOMImplementationLS) registry.getDOMImplementation("LS");
            LSParser lsParser = domLS.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);

            LSSerializer lsSerializer = domLS.createLSSerializer();
            DOMConfiguration config = lsSerializer.getDomConfig();
            config.setParameter("format-pretty-print", true);            
        	
            org.w3c.dom.Document doc = lsParser.parseURI(new File(fileLabel.getText()).toURI().toString());
            XPath xp = XPathFactory.newInstance().newXPath();
            xp.setNamespaceContext(new DocumentNamespaceContext(doc));
            NodeList nodes = (NodeList) xp.evaluate(xpath.getText(), doc, XPathConstants.NODESET);
            StringBuilder sb = new StringBuilder();
            if (nodes.getLength() > 0) {
                Node node = nodes.item(0);
                if (qname == null) {
                    qname = new QName(node.getNamespaceURI(), node.getLocalName()).toString();
                }
                sb.append(lsSerializer.writeToString(node));
            }
        	            
            if (show) {
                MessageBox mbx = new MessageBox(getSite().getShell());
                mbx.setMessage(sb.toString());
                mbx.setText(qname != null ? qname : "");
                mbx.open();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return qname;
    }
    
    private static class MyPolicyDetailEditorInput extends FileEditorInput implements IPolicyDetailEditorInput {
        private final IProject prj;
        private final ISchemaProvider schemaProvider;        
        private final InputStream inputStream;
        private HashMap<String, Object> properties = new HashMap<String, Object>();       

        public MyPolicyDetailEditorInput(IFile file, ISchemaProvider sp, InputStream is) {
            super(file);
            
            prj = file.getProject();
            schemaProvider = sp;
            inputStream = is;
        }

        public IPolicyDetailCallback getCallback() {
            return null;
        }

        public IProject getProject() {
            return prj;
        }

        public Map<String, Object> getProperties() {
            return properties;
        }

        public ISchemaProvider getSchemaProvider() {
            return schemaProvider;
        }

        public InputStream getStream() {
            return inputStream;
        }
        
    }
}